@@ -17,19 +17,105 @@
 framework and the materialization engine."""
 
 import navis
+import requests
 
 import datetime as dt
 import numpy as np
 import pandas as pd
 
-from .utils import get_cave_client, retry
-from .segmentation import locs_to_segments
+from requests_futures.sessions import FuturesSession
+
+from ..utils import make_iterable
+from .utils import get_cave_client, retry, get_chunkedgraph_secret
+from .segmentation import locs_to_segments, supervoxels_to_roots, is_latest_root
 
 
 __all__ = ['get_somas', 'get_materialization_versions',
            'create_annotation_table', 'get_annotation_tables',
            'get_annotation_table_info', 'get_annotations',
-           'delete_annotations', 'upload_annotations']
+           'delete_annotations', 'upload_annotations',
+           'is_proofread']
+
+
+PR_TABLE = None
+PR_META = None
+
+
+def is_proofread(x, cache=True):
+    """Test if neuron has been set to `proofread`.
+
+    Under the hood, this uses a cached version of the proofreading table which
+    is updated everytime this function gets called.
+
+    Parameters
+    ----------
+    x  :            int | list of int
+                    Root IDs to check.
+    cache :         bool
+                    Use and update a locally cached version of the proofreading
+                    table. Setting this to ``False`` will force fetching the
+                    full table which is considerably slower.
+
+    Returns
+    -------
+    proofread :     np.ndarray
+                    Boolean array.
+
+    """
+    global PR_TABLE, PR_META
+
+    if not isinstance(x, (np.ndarray, set, list, pd.Series)):
+        x = [x]
+
+    # Force into array and convert to integer
+    x = np.asarray(x).astype(np.int64)
+
+    # Check if any of the roots are outdated -> can't check those
+    il = is_latest_root(x)
+    if any(~il):
+        print("At least some root ID(s) outdated and will therefore show up as "
+              f"not proofread: {x[~il]}")
+
+    # Get available materialization versions
+    client = get_cave_client('production')
+    mat_versions = client.materialize.get_versions()
+
+    # Check if the cached version is outdated
+    if cache and PR_META:
+        if max(mat_versions) > PR_META['mat_version']:
+            PR_TABLE = None
+            PR_META = None
+
+    # If nothing cached catch the table from scratch
+    if isinstance(PR_TABLE, type(None)) or not cache:
+        # This ought to automatically use the most recent materialization version
+        PR_META = dict(timestamp=dt.datetime.utcnow(),
+                       mat_version=max(mat_versions))
+        PR_TABLE = client.materialize.live_query(table='proofreading_status_public_v1',
+                                                 timestamp=PR_META['timestamp'])
+        # Only keep relevant rows and columns
+        PR_TABLE = PR_TABLE.loc[PR_TABLE.valid == 't', ['pt_supervoxel_id', 'pt_root_id']]
+    else:
+        # Update root IDs
+        now = dt.datetime.utcnow()
+        # get_delta_roots currently acts up if there are no new roots
+        # (i.e. if this gets called in quick succession)
+        try:
+            old_roots, _ = client.chunkedgraph.get_delta_roots(PR_META['timestamp'], now)
+        except requests.exceptions.HTTPError as e:
+            if "need at least one array" in str(e):
+                old_roots = []
+            else:
+                raise
+        except BaseException:
+            raise
+
+        to_update = PR_TABLE.pt_root_id.isin(old_roots)
+        if any(to_update):
+            PR_TABLE.loc[to_update, 'pt_root_id'] = supervoxels_to_roots(PR_TABLE.loc[to_update, 'pt_supervoxel_id'].values)
+        PR_META['timestamp'] = now
+
+    return np.isin(x, PR_TABLE.pt_root_id.values)
 
 
 def get_materialization_versions(dataset='production'):
@@ -167,9 +253,9 @@ def get_annotation_table_info(table_name: str,
 
 def get_annotations(table_name: str,
                     materialization='live',
-                    split_positions=False,
+                    split_positions: bool = False,
                     drop_invalid: bool = True,
-                    dataset='production',
+                    dataset: str = 'production',
                     **filters):
     """Get annotations from given table.
 
@@ -384,9 +470,9 @@ def get_somas(x=None, materialization='live', split_positions=False, dataset='pr
     filter_in_dict = None
     if not isinstance(x, type(None)):
         if isinstance(x, navis.NeuronList):
-            root_ids = x.id.astype(int)
+            root_ids = x.id.astype(np.int64)
         else:
-            root_ids = navis.utils.make_iterable(x).astype(int)
+            root_ids = make_iterable(x, force_type=np.int64)
         filter_in_dict = {'pt_root_id': root_ids}
 
     nuc = get_annotations('nuclei_v1',
@@ -433,3 +519,232 @@ def get_somas(x=None, materialization='live', split_positions=False, dataset='pr
     # Sorting by radius makes sure that the small false-positives end up at the
     # bottom of the list
     return nuc.sort_values('rad_est', ascending=False)
+
+
+def submit_cell_identification(x, split_tags=False, validate=True, max_threads=4, progress=True):
+    """Submit a identification for given cells.
+
+    Use this bulk submission of cell identification with great care!
+
+    Parameters
+    ----------
+    x :             pandas.DataFrame
+                    Must have the following columns:
+                      - `valid_id` contains the current root ID
+                      - `x`, `y`, `z` (or alternatively `pos_x`, `pos_y`,
+                        `pos_z`) contain coordinates mapping to that root
+                        (must be in voxel space)
+                      - `tags` must be a comma-separated string of tags
+                      - `user_id` (optional) if you want to submit for someone
+                        else
+    split_tags :    bool
+                    If True, will split the comma-separated tags into
+                    individual tags.
+    max_threads :   int
+                    Number of parallel submissions.
+    progress :      bool
+                    If True, show progress bar.
+
+    Returns
+    -------
+    response :      pandas.DataFrame
+                    A copy of ``x`` with a `success` and an `error` column.
+
+    """
+    if not isinstance(x, pd.DataFrame):
+        raise TypeError(f'Expected DataFrame, got {type(x)}')
+
+    REQ_COLS = ('valid_id',
+                ('x', 'pos_x'),
+                ('y', 'pos_y'),
+                ('z', 'pos_z'),
+                'tags')
+    for c in REQ_COLS:
+        if isinstance(c, tuple):
+            # Check that at least one option exits
+            if not any(np.isin(c, x.columns)):
+                raise ValueError(f'`x` must contain one of these column: {c}')
+            # Rename so we always find the first possible option
+            if c[0] not in x.columns:
+                for v in c[1:]:
+                    if v in x.columns:
+                        x = x.rename({v: c[0]}, axis=1)
+        else:
+            if c not in x:
+                raise ValueError(f'Missing required column: {c}')
+
+    if validate:
+        roots = locs_to_segments(x[['x', 'y', 'z']].values)
+
+        is_zero = roots == 0
+        if any(is_zero):
+            raise ValueError(f'{is_zero.sum()} xyz coordinates map to root ID 0')
+
+        mm = roots != x.valid_id.astype(int)
+        if any(mm):
+            raise ValueError(f'{mm.sum()} xyz coordinates do not map to the '
+                             'provided `valid_id`')
+
+    session = requests.Session()
+    future_session = FuturesSession(session=session, max_workers=max_threads)
+
+    token = get_chunkedgraph_secret()
+    session.headers['Authorization'] = f"Bearer {token}"
+
+    futures = {}
+    url = f'https://prod.flywire-daf.com/neurons/api/v1/submit_cell_identification'
+    for i, row in x.iterrows():
+        if split_tags:
+            tags = row.tag.split(',')
+        else:
+            tags = [row.tag]
+
+        for tag in tags:
+            post = dict(valid_id=str(row.valid_id),
+                        location=f'{row.x}, {row.y}, {row.z}',
+                        tag=tag,
+                        action='single',
+                        user_id=row.get('user_id', ''))
+
+            f = future_session.post(url, data=post)
+            futures[f] = (i, post)
+
+    # Get the responses
+    resp = [f.result() for f in navis.config.tqdm(futures,
+                                                  desc='Submitting',
+                                                  disable=not progress or len(futures) == 1,
+                                                  leave=False)]
+
+    success = [True] * len(x)
+    errors = [[]] * len(x)
+    for r, f in zip(resp, futures):
+        row_ix = futures[f][0]
+        post = futures[f][1]
+        try:
+            r.raise_for_status()
+        except BaseException as e:
+            success[row_ix] = False
+            errors[row_ix].append(str(e))
+            continue
+
+        if not 'Success' in r.text:
+            success[row_ix] = False
+            errors[row_ix].append(r.text)
+            continue
+
+        errors[row_ix].append(None)
+
+    x = x.copy()
+    x['success'] = success
+    x['errors'] = errors
+
+    if not all(x.success):
+        navis.config.logger.error(f'Encountered {(~x.success).sum()} errors '
+                                  'while marking cells as proofread. Please '
+                                  'see the `errors` columns in the returned '
+                                  'dataframe for details.')
+    else:
+        navis.config.logger.info('SUCCESS!')
+
+    return x
+
+
+def mark_cell_completion(x, validate=True, max_threads=4, progress=True):
+    """Submit proofread status for given cell.
+
+    Use this bulk submission of proofreading status with great care!
+
+    Parameters
+    ----------
+    x :             pandas.DataFrame
+                    Must have the following columns:
+                      - `valid_id` contains the current root ID
+                      - `x`, `y`, `z` contain coordinates mapping to that root
+                        (must be in voxel space)
+                      - `user_id` (optional) if you want to submit for someone
+                        else
+    max_threads :   int
+                    Number of parallel submissions.
+    progress :      bool
+                    If True, show progress bar.
+
+    Returns
+    -------
+    response :      pandas.DataFrame
+                    A copy of ``x`` with a `success` and an `error` column.
+
+    """
+    if not isinstance(x, pd.DataFrame):
+        raise TypeError(f'Expected DataFrame, got {type(x)}')
+
+    REQ_COLS = ('valid_id', 'x' , 'y' , 'z')
+    for c in REQ_COLS:
+        if c not in x:
+            raise ValueError(f'Missing required column: {c}')
+
+    if validate:
+        roots = locs_to_segments(x[['x', 'y', 'z']].values)
+
+        is_zero = roots == 0
+        if any(is_zero):
+            raise ValueError(f'{is_zero.sum()} xyz coordinates map to root ID 0')
+
+        mm = roots != x.valid_id.astype(int)
+        if any(mm):
+            raise ValueError(f'{mm.sum()} xyz coordinates do not map to the '
+                             'provided `valid_id`')
+
+    session = requests.Session()
+    future_session = FuturesSession(session=session, max_workers=max_threads)
+
+    token = get_chunkedgraph_secret()
+    session.headers['Authorization'] = f"Bearer {token}"
+
+    futures = {}
+    url = f'https://prod.flywire-daf.com/neurons/api/v1/mark_completion'
+    for i, row in x.iterrows():
+        post = dict(valid_id=str(row.valid_id),
+                    location=f'{row.x}, {row.y}, {row.z}',
+                    action='single',
+                    user_id=row.get('user_id', ''))
+
+        f = future_session.post(url, data=post)
+        futures[f] = post
+
+    # Get the responses
+    resp = [f.result() for f in navis.config.tqdm(futures,
+                                                  desc='Submitting',
+                                                  disable=not progress or len(futures) == 1,
+                                                  leave=False)]
+
+    success = []
+    errors = []
+    for r, f in zip(resp, futures):
+        try:
+            r.raise_for_status()
+        except BaseException as e:
+            sucess.append(False)
+            errors.append(str(e))
+            continue
+
+        if not 'Success' in r.text:
+            success.append(False)
+            errors.append(r.text)
+            continue
+
+        success.append(True)
+        errors.append(None)
+
+    x = x.copy()
+    x['success'] = success
+    x['errors'] = errors
+
+    if not all(x.success):
+        navis.config.logger.error(f'Encountered {(~x.success).sum()} errors '
+                                  'while submitting cell identifications. Please '
+                                  'see the `errors` columns in the returned '
+                                  'dataframe for details.')
+    else:
+        navis.config.logger.info('SUCCESS!')
+
+    return x
