@@ -74,7 +74,7 @@ def l2_info(root_ids, progress=True, max_threads=4, dataset='production'):
         root_ids = np.unique(root_ids)
         info = []
         with ThreadPoolExecutor(max_workers=max_threads) as pool:
-            func = retry(partial(l2_info, dataset=dataset))
+            func = partial(l2_info, dataset=dataset)
             futures = pool.map(func, root_ids)
             info = [f for f in navis.config.tqdm(futures,
                                                  desc='Fetching L2 info',
@@ -90,7 +90,8 @@ def l2_info(root_ids, progress=True, max_threads=4, dataset='production'):
     l2_ids = get_l2_ids(root_ids)
 
     attributes = ['area_nm2', 'size_nm3', 'max_dt_nm', 'rep_coord_nm']
-    info = client.l2cache.get_l2data(l2_ids.tolist(), attributes=attributes)
+    get_l2data = retry(client.l2cache.get_l2data)
+    info = get_l2data(l2_ids.tolist(), attributes=attributes)
     n_miss = len([v for v in info.values() if not v])
 
     row = [root_ids, len(l2_ids), n_miss]
@@ -99,7 +100,7 @@ def l2_info(root_ids, progress=True, max_threads=4, dataset='production'):
 
     # Collect L2 attributes
     for at in attributes:
-        if at in ('rep_coord_nm'):
+        if at in ('rep_coord_nm', ):
             continue
 
         summed = sum([v.get(at, 0) for v in info.values()])
@@ -189,7 +190,7 @@ def l2_graph(root_ids, progress=True, dataset='production'):
 
 
 def l2_skeleton(root_id, refine=True, drop_missing=True, omit_failures=None,
-                progress=True, dataset='production', **kwargs):
+                progress=True, max_threads=4, dataset='production', **kwargs):
     """Generate skeleton from L2 graph.
 
     Parameters
@@ -216,6 +217,9 @@ def l2_skeleton(root_id, refine=True, drop_missing=True, omit_failures=None,
                          - ``False`` will return an empty ``TreeNeuron``
     progress :          bool
                         Whether to show a progress bar.
+    max_threads :       int
+                        Number of parallel requests to make when fetching the
+                        L2 skeletons.
     **kwargs
                         Keyword arguments are passed through to Dotprops
                         initialization. Use to e.g. set extra properties.
@@ -242,18 +246,24 @@ def l2_skeleton(root_id, refine=True, drop_missing=True, omit_failures=None,
     if omit_failures not in (None, True, False):
         raise ValueError('`omit_failures` must be either None, True or False. '
                          f'Got "{omit_failures}".')
-    # TODO:
-    # - drop duplicate nodes in unrefined skeleton
-    # - use L2 graph to find soma: highest degree is typically the soma
 
     if navis.utils.is_iterable(root_id):
-        nl = []
-        for id in navis.config.tqdm(root_id, desc='L2 skeletons',
-                                    disable=not progress, leave=False):
-            n = l2_skeleton(id, refine=refine, drop_missing=drop_missing,
-                            omit_failures=omit_failures,
-                            progress=progress, dataset=dataset, **kwargs)
-            nl.append(n)
+        get_l2_skels = partial(l2_skeleton, refine=refine, drop_missing=drop_missing,
+                               omit_failures=omit_failures, dataset=dataset, **kwargs)
+        if (max_threads > 1) and (len(root_id) > 1):
+            with ThreadPoolExecutor(max_workers=max_threads) as pool:
+                futures = pool.map(get_l2_skels, root_id)
+                nl = [f for f in navis.config.tqdm(futures,
+                                                   desc='Fetching L2 skeletons',
+                                                   total=len(root_id),
+                                                   disable=not progress or len(root_id) == 1,
+                                                   leave=False)]
+        else:
+            nl = [get_l2_skels(r) for r in navis.config.tqdm(root_id,
+                                               desc='Fetching L2 skeletons',
+                                               total=len(root_id),
+                                               disable=not progress or len(root_id) == 1,
+                                               leave=False)]
         return navis.NeuronList(nl)
 
     # Turn into integer
@@ -322,9 +332,11 @@ def l2_skeleton(root_id, refine=True, drop_missing=True, omit_failures=None,
 
     if refine:
         # Get the L2 representative coordinates
-        l2_info = client.l2cache.get_l2data(l2_ids.tolist(), attributes=['rep_coord_nm'])
+        get_l2data = retry(client.l2cache.get_l2data)
+        l2_info = get_l2data(l2_ids.tolist(), attributes=['rep_coord_nm', 'max_dt_nm'])
         # Missing L2 chunks will be {'id': {}}
         new_co = {l2dict[int(k)]: v['rep_coord_nm'] for k, v in l2_info.items() if v}
+        new_r = {l2dict[int(k)]: v.get('max_dt_nm', 0) for k, v in l2_info.items() if v}
 
         # Map refined coordinates onto the SWC
         has_new = swc.node_id.isin(new_co)
@@ -336,6 +348,8 @@ def l2_skeleton(root_id, refine=True, drop_missing=True, omit_failures=None,
             swc.loc[has_new, 'y'] = swc.loc[has_new, 'node_id'].map(lambda x: new_co[x][1])
             swc.loc[has_new, 'z'] = swc.loc[has_new, 'node_id'].map(lambda x: new_co[x][2])
 
+        swc['radius'] = swc.node_id.map(new_r)
+
         # Turn into a proper neuron
         tn = navis.TreeNeuron(swc, id=root_id, units='1 nm', **kwargs)
 
@@ -360,11 +374,12 @@ def l2_skeleton(root_id, refine=True, drop_missing=True, omit_failures=None,
     return tn
 
 
-def l2_dotprops(root_ids, min_size=None, omit_failures=None, progress=True,
-                max_threads=10, dataset='production', **kwargs):
+def l2_dotprops(root_ids, min_size=None, sample=False, omit_failures=None,
+                progress=True, max_threads=10, dataset='production', **kwargs):
     """Generate dotprops from L2 chunks.
 
-    L2 chunks not present in the L2 cache are silently ignored.
+    L2 chunks not present in the L2 cache or without a `pca` attribute
+    (happens for very small chunks) are silently ignored.
 
     Parameters
     ----------
@@ -377,6 +392,10 @@ def l2_dotprops(root_ids, min_size=None, omit_failures=None, progress=True,
                         finer terminal neurites which typically break into more,
                         smaller chunks and are hence overrepresented. A good
                         value appears to be around 1_000_000.
+    sample :            float [0 > 1], optional
+                        If float, will create Dotprops based on a fractional
+                        sample of the L2 chunks. The sampling is random but
+                        deterministic.
     omit_failures :     bool, optional
                         Determine behaviour when dotprops generation fails
                         (i.e. if the neuron has no L2 info):
@@ -419,7 +438,7 @@ def l2_dotprops(root_ids, min_size=None, omit_failures=None, progress=True,
     if not navis.utils.is_iterable(root_ids):
         root_ids = [root_ids]
 
-    root_ids = np.asarray(root_ids)
+    root_ids = np.asarray(root_ids, dtype=np.int64)
 
     if '0' in root_ids or 0 in root_ids:
         raise ValueError('Unable to produce dotprops for root ID 0.')
@@ -440,6 +459,17 @@ def l2_dotprops(root_ids, min_size=None, omit_failures=None, progress=True,
     # Turn IDs into strings
     l2_ids = [i.astype(str) for i in l2_ids]
 
+    if sample:
+        if sample <= 0 or sample >= 1:
+            raise ValueError(f'`sample` must be between 0 and 1, got {sample}')
+
+        for i in range(len(l2_ids)):
+            # Make the sampling deterministic
+            np.random.seed(1985)
+            l2_ids[i] = np.random.choice(l2_ids[i],
+                                         size=max(1, int(len(l2_ids[i]) * sample)),
+                                         replace=False)
+
     # Flatten into a list of all L2 IDs
     l2_ids_all = np.unique([i for l in l2_ids for i in l])
 
@@ -454,15 +484,16 @@ def l2_dotprops(root_ids, min_size=None, omit_failures=None, progress=True,
                            disable=not progress,
                            total=len(l2_ids_all),
                            leave=False) as pbar:
-        func = retry(client.l2cache.get_l2data)
+        get_l2data = retry(client.l2cache.get_l2data)
         for chunk_ix in np.arange(0, len(l2_ids_all), chunk_size):
             chunk = l2_ids_all[chunk_ix: chunk_ix + chunk_size]
-            l2_info.update(func(chunk.tolist(), attributes=attributes))
+            l2_info.update(get_l2data(chunk.tolist(), attributes=attributes))
             pbar.update(len(chunk))
 
     # L2 chunks without info will show as empty dictionaries
     # Let's drop them to make our life easier (speeds up indexing too)
-    l2_info = {k: v for k, v in l2_info.items() if v}
+    # Note that small L2 chunks won't have a `pca`
+    l2_info = {k: v for k, v in l2_info.items() if 'pca' in v}
 
     # Generate dotprops
     dps = []
