@@ -19,9 +19,11 @@ import requests
 import textwrap
 import time
 
+import cloudvolume as cv
 import datetime as dt
 import numpy as np
 import pandas as pd
+import networkx as nx
 
 from concurrent import futures
 from diskcache import Cache
@@ -32,15 +34,121 @@ from tqdm.auto import trange, tqdm
 from .. import spine
 from .. import xform
 
+from ..utils import make_iterable
 from .utils import (parse_volume, FLYWIRE_DATASETS, get_chunkedgraph_secret,
-                    retry, get_cave_client)
+                    retry, get_cave_client, parse_bounds)
 
 
 __all__ = ['fetch_edit_history', 'fetch_leaderboard', 'locs_to_segments',
            'locs_to_supervoxels', 'skid_to_id', 'update_ids',
            'roots_to_supervoxels', 'supervoxels_to_roots',
            'neuron_to_segments', 'is_latest_root', 'is_valid_root',
-           'is_valid_supervoxel']
+           'is_valid_supervoxel', 'get_voxels', 'get_lineage_graph']
+
+
+def get_lineage_graph(x, size=False, user=False, synapses=False,
+                      proofreading_status=False, progress=True,
+                      dataset='production'):
+    """Get lineage graph for given neuron.
+
+    This piggy-backs on the CAVEclient but importantly we remap users and
+    operation IDs such that each node's labels refer to the operation that
+    created them.
+
+    Parameters
+    ----------
+    x :         int
+                A single root ID.
+    size :      bool
+                If True, will add `size` and `survivals` node attributes. The
+                former indicates the number of supervoxels, the latter how many
+                of these supervoxels made it into `x`.
+    synapses :  bool
+                If True, will add `pre|post|synapses` node attributes which
+                indicate how many of the synapses in `x` came from this fragment.
+                Note that this doesn't tell you e.g. how many false-positive
+                synapses were removed via a split. This works only if the root
+                ID is up-to-date.
+    user :      bool
+                If True, will add user `user` node attribute.
+    proofreading_status : bool
+                If True, will add a `proofread_by` node attribute indicating if
+                a user has set a given root ID to proofread.
+
+    Returns
+    -------
+    networkx.DiGraph
+
+    """
+    x = int(x)
+
+    client = get_cave_client(dataset=dataset)
+    G = client.chunkedgraph.get_lineage_graph(x, as_nx_graph=True)
+
+    # Remap operation ID
+    op_remapped = {}
+    for n in G:
+        pred = list(G.predecessors(n))
+        if pred:
+            op_remapped[n] = G.nodes[pred[0]]['operation_id']
+
+    # Remove existing operation IDs
+    for n in G.nodes:
+        G.nodes[n].pop('operation_id', None)
+    # Apply new IDs
+    nx.set_node_attributes(G, op_remapped, name='operation_id')
+
+    if user:
+        op_ids = nx.get_node_attributes(G, "operation_id")
+        details = client.chunkedgraph.get_operation_details(list(op_ids.values()))
+        users = {n: details[str(o)]['user'] for n, o in op_ids.items()}
+        nx.set_node_attributes(G, users, name='user')
+
+    if size:
+        sv = roots_to_supervoxels(list(G.nodes), dataset=dataset, progress=progress)
+        sizes = {n: len(sv[n]) for n in G.nodes}
+        nx.set_node_attributes(G, sizes, name='size')
+
+        survivors = {n: int(np.isin(sv[n], sv[x]).sum()) for n in G.nodes}
+        nx.set_node_attributes(G, survivors, name='survivors')
+    else:
+        sv = None
+
+    if synapses:
+        pre = client.materialize.live_query(
+                        table=client.materialize.synapse_table,
+                        filter_equal_dict=dict(pre_pt_root_id=x),
+                        timestamp=dt.datetime.now(),
+                        select_columns=['pre_pt_supervoxel_id',
+                                        'post_pt_supervoxel_id']
+                                        )
+        post = client.materialize.live_query(
+                        table=client.materialize.synapse_table,
+                        filter_equal_dict=dict(post_pt_root_id=x),
+                        timestamp=dt.datetime.now(),
+                        select_columns=['pre_pt_supervoxel_id',
+                                        'post_pt_supervoxel_id']
+                                        )
+        if isinstance(sv, type(None)):
+            sv = roots_to_supervoxels(list(G.nodes), dataset=dataset, progress=progress)
+
+        n_pre = {n: int(pre.pre_pt_supervoxel_id.isin(sv[n]).sum()) for n in G.nodes}
+        n_post = {n: int(post.post_pt_supervoxel_id.isin(sv[n]).sum()) for n in G.nodes}
+        n_syn = {n: n_pre[n] + n_post[n] for n in G.nodes}
+        nx.set_node_attributes(G, n_pre, name='presynapses')
+        nx.set_node_attributes(G, n_post, name='postsynapses')
+        nx.set_node_attributes(G, n_post, name='synapses')
+
+    if proofreading_status:
+        from .annotations import get_annotations
+        nodes = np.array(list(G.nodes)).astype(np.int64)
+        pr = get_annotations('proofreading_status_public_v1',
+                             filter_in_dict=dict(valid_id=nodes))
+        if len(pr):
+            user = pr.groupby('valid_id').user_id.apply(list).to_dict()
+            nx.set_node_attributes(G, {n: user[n] for n in pr.valid_id}, name='proofread_by')
+
+    return G
 
 
 def fetch_leaderboard(days=7, by_day=False, progress=True, max_threads=4):
@@ -118,6 +226,9 @@ def fetch_leaderboard(days=7, by_day=False, progress=True, max_threads=4):
 def fetch_edit_history(x, dataset='production', progress=True, max_threads=4):
     """Fetch edit history for given neuron(s).
 
+    Note that neurons that haven't seen any edits will simply not show up in
+    returned table.
+
     Parameters
     ----------
     x :             int | list of int
@@ -174,6 +285,13 @@ def fetch_edit_history(x, dataset='production', progress=True, max_threads=4):
 
     df = []
     for r, i in zip(resp, x):
+        # Code 500 means server error
+        if r.status_code == 500:
+            # If server responds a time-out, it means that the root ID has not
+            # seen any edits from base segmentation.
+            if 'Read timed out' in r.json().get('message', ''):
+                continue
+
         r.raise_for_status()
         this_df = pd.DataFrame(r.json())
         this_df['segment'] = i
@@ -224,7 +342,10 @@ def roots_to_supervoxels(x, use_cache=True, dataset='production', progress=True)
 
     """
     # Make sure we are working with an array of integers
-    x = navis.utils.make_iterable(x).astype(int, copy=False)
+    x = make_iterable(x, force_type=np.int64)
+
+    # Make sure we're not getting bogged down with duplicates
+    x = np.unique(x)
 
     if len(x) <= 1:
         progress = False
@@ -233,7 +354,7 @@ def roots_to_supervoxels(x, use_cache=True, dataset='production', progress=True)
     vol = parse_volume(dataset)
 
     svoxels = {}
-    # See what ewe can get from cache
+    # See what we can get from cache
     if use_cache:
         # Cache for root -> supervoxels
         # Grows to max 1Gb by default and persists across sessions
@@ -252,14 +373,18 @@ def roots_to_supervoxels(x, use_cache=True, dataset='production', progress=True)
     # We need to convert keys to integer array because otherwise there is a
     # mismatch in types (int vs np.int?) which causes all root IDs to be in miss
     # -> I think that's because of the way disk cache works
-    miss = x[~np.isin(x, np.array(list(svoxels.keys())).astype(int))]
+    miss = x[~np.isin(x, np.array(list(svoxels.keys()), dtype=np.int64))]
     get_leaves = retry(vol.get_leaves)
-    svoxels.update({i: get_leaves(i,
-                                  bbox=vol.meta.bounds(0),
-                                  mip=0) for i in navis.config.tqdm(miss,
-                                                                    desc='Querying',
-                                                                    disable=not progress,
-                                                                    leave=False)})
+    with navis.config.tqdm(desc='Querying',
+                           total=len(x),
+                           disable=not progress,
+                           leave=False) as pbar:
+        # Update for those for which we had cached data
+        pbar.update(len(svoxels))
+
+        for i in miss:
+            svoxels[i] = get_leaves(i, bbox=vol.meta.bounds(0), mip=0)
+            pbar.update()
 
     # Update cache
     if use_cache:
@@ -270,7 +395,7 @@ def roots_to_supervoxels(x, use_cache=True, dataset='production', progress=True)
     return svoxels
 
 
-def supervoxels_to_roots(x, timestamp=None, batch_size=10_000,
+def supervoxels_to_roots(x, timestamp=None, batch_size=10_000, stop_layer=10,
                          retry=True, progress=True, dataset='production'):
     """Get root(s) for given supervoxel(s).
 
@@ -285,6 +410,8 @@ def supervoxels_to_roots(x, timestamp=None, batch_size=10_000,
     batch_size :    int
                     Max number of supervoxel IDs per query. Reduce batch size if
                     you experience time outs.
+    stop_layer :    int
+                    Set e.g. to ``2`` to get L2 IDs instead of root IDs.
     retry :         bool
                     Whether to retry if a batched query fails.
     dataset :       str | CloudVolume
@@ -306,13 +433,11 @@ def supervoxels_to_roots(x, timestamp=None, batch_size=10_000,
 
     """
     # Make sure we are working with an array of integers
-    x = navis.utils.make_iterable(x)
+    x = make_iterable(x, force_type=np.int64)
 
     # Check if IDs are valid (zeros are fine because we filter for them later on)
     is_valid_supervoxel(x[(x != 0) & (x != '0')], raise_exc=True)
 
-    x = x.astype(np.int64, copy=False)
-
     # Parse the volume
     vol = parse_volume(dataset)
 
@@ -335,12 +460,14 @@ def supervoxels_to_roots(x, timestamp=None, batch_size=10_000,
             not_zero = batch != 0
             try:
                 roots[i:i+batch_size][not_zero] = vol.get_roots(batch[not_zero],
+                                                                stop_layer=stop_layer,
                                                                 timestamp=timestamp)
             except BaseException:
                 if not retry:
                     raise
                 time.sleep(1)
                 roots[i:i+batch_size][not_zero] = vol.get_roots(batch[not_zero],
+                                                                stop_layer=stop_layer,
                                                                 timestamp=timestamp)
 
             pbar.update(len(batch))
@@ -396,7 +523,7 @@ def locs_to_supervoxels(locs, mip=2, coordinates='voxel'):
             locs = locs.astype(np.float64)
 
     return spine.transform.get_segids(locs, segmentation='flywire_190410',
-                                      coordinates=coordinates, mip=-1)
+                                      coordinates=coordinates, mip=mip)
 
 
 def neuron_to_segments(x, short=False, dataset='production', coordinates='voxel'):
@@ -651,7 +778,7 @@ def is_latest_root(id, dataset='production', **kwargs):
     """
     dataset = FLYWIRE_DATASETS.get(dataset, dataset)
 
-    id = navis.utils.make_iterable(id).astype(str)
+    id = make_iterable(id, force_type=str)
 
     # The server doesn't like being asked for zeros
     not_zero = id != '0'
@@ -771,8 +898,8 @@ def update_ids(id,
                                                                                disable=not progress or len(id) == 1)]
             res = pd.concat(res, axis=0, sort=False, ignore_index=True)
         else:
-            if (not isinstance(supervoxels, (list, set, np.ndarray))
-                or len(supervoxels) != len(id)):
+            supervoxels = np.asarray(supervoxels)
+            if len(supervoxels) != len(id):
                 raise ValueError(f'Number of supervoxels ({len(supervoxels)}) does '
                                  f'not match number of root IDs ({len(id)})')
             elif any(pd.isnull(supervoxels)):
@@ -780,7 +907,7 @@ def update_ids(id,
             elif any(pd.isnull(id)):
                 raise ValueError('`id` must not contain `None`')
 
-            id = np.array(id).astype(int)
+            id = np.array(id).astype(np.int64)
 
             res = pd.DataFrame()
             res['old_id'] = id
@@ -835,7 +962,7 @@ def update_ids(id,
             if sample >= 1:
                 smpl = svoxels[: sample]
             else:
-                smpl = svoxels[: int(len(svoxels) * sample)]
+                smpl = svoxels[: max(int(len(svoxels) * sample), 1)]
 
             # Fetch up-to-date root IDs for the sampled supervoxels
             roots = supervoxels_to_roots(smpl, dataset=vol, progress=False)
@@ -1155,3 +1282,180 @@ def is_valid_supervoxel(x, raise_exc=False, dataset='production'):
         raise ValueError(f'{x} is not a valid supervoxel ID')
 
     return is_valid
+
+
+def get_voxels(x, mip=0, sv_map=False, bounds=None, thin=False, progress=True,
+               use_mirror=True, dataset='production'):
+    """Fetch voxels making a up given root ID.
+
+    Parameters
+    ----------
+    x :             int
+                    A single root ID.
+    mip :           int
+                    Scale at which to fetch voxels. Mip 0, for example, is
+                    16 x 16 x 40nm.
+    sv_map :        bool
+                    If True, additionally return a map with the L2 ID for each
+                    voxel.
+    bounds :        (3, 2) or (2, 3) array, optional
+                    Bounding box to return voxels in. Expected to be in 4, 4, 40
+                    voxel space.
+    thin :          bool
+                    If True, will remove voxels at the interface of adjacent
+                    supervoxels that are not supposed to be connected according
+                    to the L2 graph. This is rather expensive but can help in
+                    situations where a neuron self-touches.
+    use_mirror :    bool
+                    If True (default), will use an mirror of the base
+                    segmentation for supervoxel look-up. Possibly slightly
+                    slower than the production data set but doesn't incur
+                    egress charges for Princeton.
+    progress :      bool
+                    Whether to show a progress bar or not.
+
+    Returns
+    -------
+    voxels :        (N, 3) np.ndarray
+                    In voxel space according to `mip`.
+    sv_map :        (N, ) np.ndarray
+                    Supervoxel ID for each voxel. Only if `sv_map=True`.
+
+    """
+    # IDEA:
+    # 1. Find surface voxels for each L2 chunk
+    # 2. Get L2 graph and determine which L2 chunks are supposed to be connected
+    # 3. Remove surface voxel between adjacent but not connected L2 chunks
+
+    from .l2 import chunks_to_nm
+
+    # This is a mirror for base segmentation
+    vol = parse_volume(dataset)
+    client = get_cave_client()
+
+    if use_mirror:
+        sv_vol = cv.CloudVolume('precomputed://https://seungdata.princeton.edu/'
+                                'sseung-archive/fafbv14-ws/'
+                                'ws_190410_FAFB_v02_ws_size_threshold_200',
+                                 use_https=True, progress=False, fill_missing=True)
+    else:
+        sv_vol = vol
+
+    is_valid_root(x, raise_exc=True)
+
+    # Get L2 chunks making up this neuron
+    l2_ids = client.chunkedgraph.get_leaves(x, stop_layer=2)
+
+    # Get supervoxels for this neuron
+    sv = roots_to_supervoxels(x, dataset=dataset)[x]
+
+    # Turn l2_ids into chunk indices
+    l2_ix = [np.array(vol.mesh.meta.meta.decode_chunk_position(l)) for l in l2_ids]
+    l2_ix = np.unique(l2_ix, axis=0)
+
+    # Convert to nm
+    l2_nm = np.asarray(chunks_to_nm(l2_ix, vol=vol))
+
+    # Convert to voxel space
+    l2_vxl = l2_nm // vol.meta.scales[mip]["resolution"]
+
+    # Apply bounds
+    bounds = parse_bounds(bounds)
+    if not isinstance(bounds, type(None)):
+        base_to_mip = np.array(vol.meta.scales[mip]["resolution"]) / [4, 4, 40]
+        bounds = bounds // base_to_mip.reshape(-1, 1)
+        l2_vxl = l2_vxl[np.all(l2_vxl >= bounds[:, 0], axis=1)]
+        l2_vxl = l2_vxl[np.all(l2_vxl <= bounds[:, 1], axis=1)]
+
+    voxels = []
+    svids = []
+    ch_size = np.array(vol.mesh.meta.meta.graph_chunk_size)
+    ch_size = ch_size // (vol.mip_resolution(mip) / vol.mip_resolution(0))
+    old_mip = sv_vol.mip
+    try:
+        sv_vol.mip = mip
+        for ch in tqdm(l2_vxl,
+                       disable=not progress,
+                       leave=False,
+                       desc='Loading'):
+            ct = sv_vol[ch[0]:ch[0] + ch_size[0],
+                        ch[1]:ch[1] + ch_size[1],
+                        ch[2]:ch[2] + ch_size[2]][:, :, :, 0]
+            is_root = np.isin(ct, sv)
+            this_vxl = np.dstack(np.where(is_root))[0]
+            this_vxl = this_vxl + ch
+            voxels.append(this_vxl)
+
+            if sv_map or thin:
+                svids.append(ct[is_root])
+    except BaseException:
+        raise
+    finally:
+        sv_vol.mip = old_mip
+
+    # uint 16 should be sufficient because even at mip 0 the volume has
+    # shape (54100, 28160, 7046) -> doesn't exceed 65_535
+    voxels = np.vstack(voxels).astype('uint16')
+    if len(svids):
+        svids = np.concatenate(svids)
+
+    if thin:
+        from .l2 import l2_graph
+
+        try:
+            from pykdtree.kdtree import KDTree
+        except ImportError:
+            from scipy.spatial import cKDTree as KDTree
+
+        # Get the l2 ID for each supervoxel
+        l2_ids = vol.get_roots(svids, stop_layer=2)
+        l2_dict = dict(zip(svids, l2_ids))
+
+        # Get the l2 graph
+        G = l2_graph(x)
+
+        # Create KD tree for all voxels
+        tree = KDTree(voxels)
+
+        # Create a mask for invalidated voxels
+        invalid = np.zeros(len(voxels), dtype=bool)
+
+        # Now go over each supervoxel
+        for sv in tqdm(np.unique(svids),
+                       disable= not progress,
+                       desc='Thinning',
+                       leave=False):
+            # Get the voxels for this supervoxel
+            is_this_sv = svids == sv
+
+            # If supervoxel has no voxels just continue
+            if not np.any(is_this_sv):
+                continue
+
+            # Get all supervoxels that could be connected to this supervoxel
+            is_this_l2 = l2_ids == l2_dict[sv]
+            is_connected_l2 = np.isin(l2_ids, list(G.neighbors(l2_dict[sv])))
+
+            # The mask needs to exclude anything that:
+            # Isn't this supervoxel OR is supposed to be connected OR
+            # has already been invalidated in a prior run
+            mask = is_this_l2 | is_connected_l2 | invalid
+
+            # Find "other" voxels that touch voxels for this supervoxel
+            dist, ix = tree.query(voxels[is_this_sv],
+                                  mask=mask,
+                                  distance_upper_bound=1.75)
+            is_touching = dist < np.inf
+
+            if not np.any(is_touching):
+                continue
+
+            invalid[np.where(is_this_sv)[0][is_touching]] = True
+
+        voxels = voxels[~invalid]
+        svids = svids[~invalid]
+
+    if not sv_map:
+        return voxels
+    else:
+        return voxels, svids
