@@ -26,6 +26,7 @@ from tqdm.auto import trange
 from .segmentation import is_latest_root
 from .utils import parse_root_ids, get_cave_client, retry, get_synapse_areas
 
+from ..utils import make_iterable
 from ..synapses.utils import catmaid_table
 from ..synapses.transmitters import collapse_nt_predictions
 
@@ -84,6 +85,7 @@ def synapse_counts(x, by_neuropil=False, min_score=30, live_query=True,
                          min_score=min_score,
                          transmitters=True, live_query=live_query,
                          neuropils=by_neuropil,
+                         batch_size=batch_size,
                          dataset=dataset, **kwargs)
 
     pre = syn[syn.pre.isin(x)]
@@ -123,8 +125,7 @@ def predict_transmitter(x, single_pred=False, weighted=True, live_query=True,
                         neuropils=None, batch_size=10, dataset='production', **kwargs):
     """Fetch neurotransmitter predictions for neurons.
 
-    Based on Eckstein et al. (2020). Uses a service on services.itanna.io hosted
-    by Eric Perlman and Davi Bock. The per-synapse predictions are collapsed
+    Based on Eckstein et al. (2020). The per-synapse predictions are collapsed
     into per-neuron prediction by calculating the average confidence for
     each neurotransmitter across all synapses weighted by the "cleft score".
     Bottom line: higher confidence synapses have more weight than low confidence
@@ -185,7 +186,7 @@ def predict_transmitter(x, single_pred=False, weighted=True, live_query=True,
                          dataset=dataset, **kwargs)
 
     if not isinstance(neuropils, type(None)):
-        neuropils = navis.utils.make_iterable(neuropils)
+        neuropils = make_iterable(neuropils)
         filter_in = [n for n in neuropils if not n.startswith('~')]
         filter_out = [n[1:] for n in neuropils if n.startswith('~')]
 
@@ -304,8 +305,12 @@ def fetch_synapses(x, pre=True, post=True, attach=True, min_score=30, clean=True
         if pre:
             syn.append(func(filter_in_dict=dict(pre_pt_root_id=batch)))
 
+    # Drop attrs to avoid issues when concatenating
+    for df in syn:
+        df.attrs = {}
+
     # Combine results from batches
-    syn = pd.concat(syn, axis=0, ignore_index=True)
+    syn = pd.concat([s for s in syn if not s.empty], axis=0, ignore_index=True)
 
     # Depending on how queries were batched, we need to drop duplicate synapses
     syn.drop_duplicates('id', inplace=True)
@@ -357,23 +362,26 @@ def fetch_synapses(x, pre=True, post=True, attach=True, min_score=30, clean=True
     if attach and isinstance(x, navis.NeuronList):
         for n in x:
             presyn = postsyn = pd.DataFrame([])
+            add_cols = ['neuropil'] if neuropils else []
             if pre:
-                presyn = syn.loc[syn.pre == int(n.id),
-                                 ['pre_x', 'pre_y', 'pre_z',
-                                  'cleft_score', 'post']].rename({'pre_x': 'x',
-                                                                  'pre_y': 'y',
-                                                                  'pre_z': 'z',
-                                                                  'post': 'partner_id'},
-                                                                 axis=1)
+                cols = ['pre_x', 'pre_y', 'pre_z',
+                        'cleft_score', 'post'] + add_cols
+                presyn = syn.loc[syn.pre == int(n.id), cols
+                                 ].rename({'pre_x': 'x',
+                                           'pre_y': 'y',
+                                           'pre_z': 'z',
+                                           'post': 'partner_id'},
+                                          axis=1)
                 presyn['type'] = 'pre'
             if post:
-                postsyn = syn.loc[syn.post == int(n.id),
-                                  ['post_x', 'post_y', 'post_z',
-                                   'cleft_score', 'pre']].rename({'post_x': 'x',
-                                                                  'post_y': 'y',
-                                                                  'post_z': 'z',
-                                                                  'pre': 'partner_id'},
-                                                                 axis=1)
+                cols = ['post_x', 'post_y', 'post_z',
+                        'cleft_score', 'pre'] + add_cols
+                postsyn = syn.loc[syn.post == int(n.id), cols
+                                  ].rename({'post_x': 'x',
+                                            'post_y': 'y',
+                                            'post_z': 'z',
+                                            'pre': 'partner_id'},
+                                           axis=1)
                 postsyn['type'] = 'post'
 
             connectors = pd.concat((presyn, postsyn), axis=0, ignore_index=True)
@@ -478,18 +486,32 @@ def fetch_adjacency(sources, targets=None, min_score=30, live_query=True,
         for k in range(0, len(targets), batch_size):
             target_batch = targets[k:k+batch_size]
 
-            syn.append(func(filter_in_dict=dict(post_pt_root_id=target_batch,
-                                                pre_pt_root_id=source_batch)))
+            this = func(filter_in_dict=dict(post_pt_root_id=target_batch,
+                                            pre_pt_root_id=source_batch))
+
+            # We need to drop the .attrs (which contain meta data from queries)
+            # Otherwise we run into issues when concatenating
+            this.attrs = {}
+
+            if not this.empty:
+                syn.append(this)
 
     # Combine results from batches
-    syn = pd.concat(syn, axis=0, ignore_index=True)
+    if len(syn):
+        syn = pd.concat(syn, axis=0, ignore_index=True)
+    else:
+        adj = pd.DataFrame(np.zeros((len(sources), len(targets))),
+                           index=sources, columns=targets)
+        adj.index.name = 'source'
+        adj.columns.name = 'target'
+        return adj
 
     # Depending on how queries were batched, we need to drop duplicate synapses
     syn.drop_duplicates('id', inplace=True)
 
     # Subset to the desired neuropils
     if not isinstance(neuropils, type(None)):
-        neuropils = navis.utils.make_iterable(neuropils)
+        neuropils = make_iterable(neuropils)
 
         if len(neuropils):
             filter_in = [n for n in neuropils if not n.startswith('~')]
@@ -633,12 +655,16 @@ def fetch_connectivity(x, clean=True, style='simple', min_score=30,
         if downstream:
             syn.append(func(filter_in_dict=dict(pre_pt_root_id=batch)))
 
+    # Drop attrs to avoid issues when concatenating
+    for df in syn:
+        df.attrs = {}
+
     # Combine results from batches
-    syn = pd.concat(syn, axis=0, ignore_index=True)
+    syn = pd.concat([s for s in syn if not s.empty], axis=0, ignore_index=True)
 
     # Subset to the desired neuropils
     if not isinstance(neuropils, type(None)):
-        neuropils = navis.utils.make_iterable(neuropils)
+        neuropils = make_iterable(neuropils)
 
         if len(neuropils):
             filter_in = [n for n in neuropils if not n.startswith('~')]
