@@ -20,20 +20,24 @@ import os
 import pytz
 import time
 import requests
+import warnings
 
 from caveclient import CAVEclient
 from pathlib import Path
 from importlib import reload
+from zipfile import ZipFile
+from io import BytesIO
 
 import cloudvolume as cv
 import datetime as dt
+import trimesh as tm
 import numpy as np
 
 from .. import utils
 
 
 __all__ = ['set_chunkedgraph_secret', 'get_chunkedgraph_secret',
-           'get_cave_client']
+           'get_cave_client', 'get_neuropil_volumes']
 
 FLYWIRE_DATASETS = {'production': 'fly_v31',
                     'sandbox': 'fly_v26'}
@@ -52,6 +56,38 @@ area_ids = None
 vol_names = None
 
 
+def get_neuropil_volumes(neuropils):
+    """Load FlyWire neuropil volumes.
+
+    These meshes were originally created by for the JFRC2 brain template
+    (for citation and details see 10.5281/zenodo.10567). Here, we transformed
+    them to FlyWire (FAFB14.1) space.
+
+    Parameters
+    ----------
+    neuropils :     str | list thereof
+                    Neuropil name(s) - e.g. 'LH_R' or ['LH_R', 'LH_L'].
+
+    Returns
+    -------
+    meshes :        single navis.Volume or list thereof
+
+    """
+    if navis.utils.is_iterable(neuropils):
+        return [get_neuropil_volumes(n) for n in neuropils]
+
+    with ZipFile(data_path / 'JFRC2NP.surf.fw.zip', 'r') as zip:
+        try:
+            f = zip.read(f'{neuropils}.stl')
+        except KeyError:
+            raise ValueError(f'No mesh for neuropil "{neuropils}"')
+
+        f = zip.read(f'{neuropils}.stl')
+        m = tm.load_mesh(BytesIO(f), file_type='stl')
+
+    return navis.Volume(m, name=neuropils)
+
+
 def get_synapse_areas(ind):
     """Lazy-load synapse areas.
 
@@ -215,11 +251,11 @@ def parse_root_ids(x):
     elif isinstance(x, (int, np.int)):
         ids = [x]
     else:
-        ids = navis.utils.make_iterable(x)
+        ids = utils.make_iterable(x, force_type=np.int64)
 
     # Make sure we are working with proper numerical IDs
     try:
-        return np.asarray(ids).astype(int)
+        return np.asarray(ids, dtype=np.int64)
     except ValueError:
         raise ValueError(f'Unable to convert given root IDs to integer: {ids}')
     except BaseException:
@@ -276,6 +312,9 @@ def parse_volume(vol, **kwargs):
 def retry(func, retries=5, cooldown=2):
     """Retry function on HTTPError.
 
+    This also suppresses UserWarnings (because we typically use this for stuff
+    like the l2 Cache).
+
     Parameters
     ----------
     cooldown :  int | float
@@ -288,13 +327,41 @@ def retry(func, retries=5, cooldown=2):
     @functools.wraps(func)
     def wrapper(*args, **kwargs):
         for i in range(1, retries + 1):
-            try:
-                return func(*args, **kwargs)
-            except requests.HTTPError:
-                if i >= retries:
+            with warnings.catch_warnings():
+                warnings.simplefilter("ignore")
+                try:
+                    return func(*args, **kwargs)
+                except requests.HTTPError:
+                    if i >= retries:
+                        raise
+                except BaseException:
                     raise
-            except BaseException:
-                raise
-            time.sleep(cooldown * i)
-    return wrapper
+                time.sleep(cooldown * i)
     return wrapper
+
+
+def parse_bounds(x):
+    """Parse bounds.
+
+    Parameters
+    ----------
+    x :     (3, 2) array | (2, 3) array | None
+
+    Returns
+    -------
+    bounds :    (3, 2) np.array
+
+    """
+    if isinstance(x, type(None)):
+        return x
+
+    x = np.asarray(x)
+
+    if not x.ndim == 2 or x.shape not in [(3, 2), (2, 3)]:
+        raise ValueError('Must provide bounding box as (3, 2) or (2, 3) array, '
+                         f'got {x.shape}')
+
+    if x.shape == (2, 3):
+        x = x.T
+
+    return np.vstack((x.min(axis=1), x.max(axis=1))).T
